// `define pure_verilog

module tinycpu
	#(parameter addrSize = 10)
	(input clk,
	 input reset,
	 inout [7:0] dataBus,
	 output reg [addrSize-1:0] addrBus,
	 output weMem);

	// CPU registers
	reg [7:0] aReg;
	reg [7:0] xReg;
	reg [5:0] spReg;
	reg [addrSize-1:0] pcReg;
	reg [addrSize-1:0] eaReg;
	reg cFlagReg, zFlagReg;
	reg [3:0] opcode;
	reg [4:0] state;
			
	// control signals		
	reg [1:0] aFunc;
	reg [1:0] xFunc;
	reg [1:0] spFunc;
	reg [1:0] pcFunc;
	reg [1:0] eaFunc;
	reg [1:0] flagFunc;
	reg [4:0] nextState;
	reg [1:0] addrSelect;
	reg [1:0] dataSelect;
	reg [7:0] dataMux;
	reg we;
	wire [7:0] addSubResult;
	wire addSubCarryOut;

	// convenience methods
	wire [2:0] mathFunc = opcode[3:1];
	wire isIndexed = opcode[0];
		
	// -- states
	localparam [4:0] sFetchOp = 5'b00000,
					 sGetAddressLo = 5'b01011,
					 sGetAddressHi = 5'b10000,
					 sAddIndex = 5'b01101,
					 sDoMath = 5'b01100,
					 sDoMathImmediate = 5'b01001,					 
					 sJump0 = 5'b00001,
					 sJump1 = 5'b01010,					 
					 sSkipByte0 = 5'b00011,
					 sSkipByte1 = 5'b00010,
					 sPullA = 5'b00100,		 
					 sPushA = 5'b00101,
					 sPullX = 5'b00110,
					 sPushX = 5'b00111,	
					 sCall0 = 5'b01110,
					 sCall1 = 5'b01111,
					 sReturn0 = 5'b10001,
					 sReturn1 = 5'b10010,
					 sLoadZFlag = 5'b01000;											

	// -- register controls
	localparam [1:0] fIncrement = 2'b00,
					 fLoadLo = 2'b00,
					 fAddSub = 2'b00,
					 fLoadZFromA = 2'b00,
	                 fDecrement = 2'b01,
	                 fLoadHi = 2'b01,
	                 fNor = 2'b01,
	                 fLoadZFromX = 2'b01,
	                 fLoad = 2'b10,
	                 fAddX = 2'b10,
	                 fLoadCZFromAddSub = 2'b10,
	                 fKeep = 2'b11;
	  
	// address selects
	localparam [1:0] eAddrPC = 2'b00,
					 eAddrEA = 2'b01,
					 eAddrSP = 2'b10;
					 
	// data selects
	localparam [1:0] eDataA = 2'b00,
					 eDataX = 2'b01,
					 eDataPCLo = 2'b10,
					 eDataPCHi = 2'b11;
					 	               
	// -- current state
	always @(posedge clk or negedge reset) begin
		if (!reset)
			state <= sFetchOp;
		else 
			state <= nextState; 
	end

	// -- opcode
	always @(posedge clk) begin
		if (state == sFetchOp)
			opcode <= dataBus[4:1]; 
	end
	
	// -- stack pointer
	// can increment or decrement
	always @(posedge clk or negedge reset) begin
		if (!reset)
			spReg <= 6'b111111;
		else if (spFunc == fIncrement)
			spReg <= spReg + 1'b1;
		else if (spFunc == fDecrement)
			spReg <= spReg - 1'b1;		
	end
	
	// -- X (index) register
	// can increment, decrement, or load	
	always @(posedge clk) begin
		if (xFunc == fIncrement) 
			xReg <= xReg + 1'b1;
		else if (xFunc == fDecrement)
			xReg <= xReg - 1'b1;
		else if (xFunc == fLoad)
			xReg <= dataBus;						
	end
	
	// -- program counter
	// can increment or load
	always @(posedge clk or negedge reset) begin
		if (!reset)
			pcReg <= 0;
		else if (pcFunc == fIncrement)
			pcReg <= pcReg + 1'b1;
		else if (pcFunc == fLoad)
			pcReg <= { dataBus[addrSize-9:0], eaReg[7:0] };					
	end

	// -- effective address register
	// can add X, load lo, or load hi
	always @(posedge clk) begin
		if (eaFunc == fLoadLo)
			eaReg[7:0] <= dataBus;
		else if (eaFunc == fLoadHi)
			eaReg[addrSize-1:8] <= dataBus[addrSize-9:0];
		else if (eaFunc == fAddX)
			eaReg <= eaReg + xReg;		
	end
	
	// -- Adder/subtractor
`ifdef pure_verilog

	assign { addSubCarryOut, addSubResult } = mathFunc[0] ? aReg + dataBus : aReg - dataBus;	
	 
`else

	lpm_add_sub	addsub0 (.dataa (aReg), .add_sub (mathFunc[0]), .datab (dataBus), .cout (addSubCarryOut), .result (addSubResult));
	
	defparam
		addsub0.lpm_direction = "UNUSED",
		addsub0.lpm_hint = "ONE_INPUT_IS_CONSTANT=NO,CIN_USED=NO",
		addsub0.lpm_representation = "UNSIGNED",
		addsub0.lpm_type = "LPM_ADD_SUB",
		addsub0.lpm_width = 8;
`endif
	
	// -- A (accumulator) register
	// can add, subtract, NOR, or load	
	always @(posedge clk) begin
		if (aFunc == fAddSub)
			aReg <= addSubResult;
		else if (aFunc == fNor)
			aReg <= ~(aReg | dataBus); 
		else if (aFunc == fLoad)
			aReg <= dataBus;		
	end
	
	// -- flags registers
	always @(posedge clk) begin
		if (flagFunc == fLoadCZFromAddSub) begin
			cFlagReg <= addSubCarryOut;
			zFlagReg <= (addSubResult == 0); 
		end 
		else if (flagFunc == fLoadZFromA) 
			zFlagReg <= (aReg == 0); 
		else if  (flagFunc == fLoadZFromX) 
			zFlagReg <= (xReg == 0); 
	end
	
	// -- address output mux
	always @* begin
		case (addrSelect)
			eAddrEA:
				addrBus <= eaReg;
			eAddrSP:
				addrBus <= { {(addrSize-6){1'b1}}, spReg }; // pad address with leading 1's
			default:
				addrBus <= pcReg;
		endcase
	end
		
	// -- data output mux
	always @* begin
		case (dataSelect)
			eDataA:
				dataMux <= aReg;
			eDataX:
				dataMux <= xReg;
			eDataPCLo:
				dataMux <= pcReg[7:0];
			eDataPCHi:
				dataMux <= { {(16-addrSize){1'b0}}, pcReg[addrSize-1:8] };
		endcase
	end
	
	assign weMem = we | clk;
	assign dataBus = !we ? dataMux : 8'bzzzzzzzz;
	
	// -- control logic: state machine
	always @* begin
		// default values for control signals
		// assign xx (don't care) whenever possible, for smaller synthesis results 
		aFunc <= fKeep;
		xFunc <= fKeep;
		spFunc <= fKeep;
		pcFunc <= fKeep;
		eaFunc <= fKeep;
		flagFunc <= fKeep;
		addrSelect <= eAddrPC;
		dataSelect <= 2'bxx;	
		we <= 1'b1;
		
		case (state)
			sFetchOp: begin
				pcFunc <= fIncrement;
				eaFunc <= 2'bxx; 
				spFunc <= (dataBus[5:2] == 4'b1100) ? fIncrement : fKeep; // pla, plx, return
				xFunc <= (dataBus[5:0] == 6'b111000) ? fIncrement : // inx
						 (dataBus[5:0] == 6'b111001) ? fDecrement : // dex
						 fKeep;
				nextState <= (!dataBus[5] & dataBus[0]) ? sDoMathImmediate : // math op, immediate mode
							(!dataBus[5] & !dataBus[0]) ? sGetAddressLo : // math op, absolute or indexed
							(dataBus[5:0] == 6'b110110) ? sJump0 : // jump
							(dataBus[5:1] == 5'b11100) ? sLoadZFlag : // inx, dex
							(dataBus[5:0] == 6'b110100) ? sPushA : // pha
							(dataBus[5:0] == 6'b110101) ? sPushX : // phx
							(dataBus[5:0] == 6'b110000) ? sPullA : // pla
							(dataBus[5:0] == 6'b110001) ? sPullX : // plx
							(dataBus[5:0] == 6'b110111) ? sCall0 : // call
							(dataBus[5:0] == 6'b110010) ? sReturn0 : // return
							(dataBus[5:4] == 2'b10 && dataBus[3] && dataBus[2] && !cFlagReg) ? sSkipByte0 : // bcs not taken
							(dataBus[5:4] == 2'b10 && !dataBus[3] && dataBus[2] && !zFlagReg) ? sSkipByte0 : // beq not taken
							(dataBus[5:4] == 2'b10 && dataBus[3] && dataBus[2] && cFlagReg) ? sJump0 : // bcs taken
							(dataBus[5:4] == 2'b10 && !dataBus[3] && dataBus[2] && zFlagReg) ? sJump0 : // beq taken	
							(dataBus[5:4] == 2'b10 && dataBus[3] && !dataBus[2] && cFlagReg) ? sSkipByte0 : // bcc not taken
							(dataBus[5:4] == 2'b10 && !dataBus[3] && !dataBus[2] && zFlagReg) ? sSkipByte0 : // bne not taken
							(dataBus[5:4] == 2'b10 && dataBus[3] && !dataBus[2] && !cFlagReg) ? sJump0 : // bcc taken
							(dataBus[5:4] == 2'b10 && !dataBus[3] && !dataBus[2] && !zFlagReg) ? sJump0 : // bne taken
							5'bxxxxx; // invalid opcode
			end
			sGetAddressLo, sJump0: begin
				// ealo <= dbus, pc <= pc+1
				eaFunc <= fLoadLo;
				pcFunc <= fIncrement;
				nextState <= (state == sJump0) ? sJump1 : sGetAddressHi;
			end			
			sGetAddressHi: begin
				// eahi <= dbus
				eaFunc <= fLoadHi;
				nextState <= isIndexed ? sAddIndex : sDoMath;
			end
			sAddIndex: begin
				// ea <= ea+x
				eaFunc <= fAddX;
				addrSelect <= 2'bxx; // not using address bus
				nextState <= sDoMath;
			end			
			sDoMath, sDoMathImmediate: begin
				// load/store/ALU, pc <= pc+1
				addrSelect <= (state == sDoMathImmediate) ? eAddrPC : eAddrEA;
				pcFunc <= fIncrement;
				eaFunc <= 2'bxx; 			
						 
				if (mathFunc[2] && mathFunc[0]) begin // sta, stx
					dataSelect <= mathFunc[1] ? eDataX : eDataA;
					we <= 1'b0;
				end else if (!mathFunc[2] && !mathFunc[1]) begin // add, sub
					aFunc <= fAddSub;
					flagFunc <= fLoadCZFromAddSub;
				end else if (mathFunc == 3'b010) // compare
					flagFunc <= fLoadCZFromAddSub; 
				else if (mathFunc == 3'b011) // nor
					aFunc <= fNor; 
				else if (mathFunc == 3'b100) // lda
					aFunc <= fLoad;
				else if (mathFunc == 3'b110) // ldx
					xFunc <= fLoad;
				
				// sta, stx, add, sub, cmp could really transition to sFetchOp, but this
				// complicates the state logic and increases the resource usage.		
				nextState <= sLoadZFlag;		
			end
			sLoadZFlag: begin
				flagFunc <= (mathFunc == 3'b011 || mathFunc == 3'b100) ? fLoadZFromA : // nor, lda
							(opcode[3:1] == 3'b110) ? fLoadZFromX : // ldx, inx, dex
							fKeep;
				eaFunc <= 2'bxx;
				addrSelect <= 2'bxx; // not using address bus
				nextState <= sFetchOp;
			end		
			sJump1: begin
				// pc <= {dbus,ealo}
				pcFunc <= fLoad;
				eaFunc <= 2'bxx;
				nextState <= sFetchOp;
			end
			sSkipByte0, sSkipByte1: begin
				// pc <= pc+1
				pcFunc <= fIncrement;
				eaFunc <= 2'bxx;
				addrSelect <= 2'bxx; // not using address bus
				nextState <= (state == sSkipByte0) ? sSkipByte1 : sFetchOp;
			end			
			sPushA, sPushX: begin
				// addrsel=sp, dbus <= a/x, sp <= sp-1 
				eaFunc <= 2'bxx;
				addrSelect <= eAddrSP;
				we <= 1'b0;
				dataSelect <= (state == sPushA) ? eDataA : eDataX;
				spFunc <= fDecrement;
				nextState <= sFetchOp;	
			end
			sPullA, sPullX: begin
				// addrsel=sp, a/x <= dbus
				eaFunc <= 2'bxx;
				addrSelect <= eAddrSP;
				aFunc <= (state == sPullA) ? fLoad : fKeep;
				xFunc <= (state == sPullX) ? fLoad : fKeep;
				nextState <= sFetchOp;
			end
			sCall0: begin
				// addrsel=sp, dbus <= pchi, sp <= sp-1 
				eaFunc <= 2'bxx;
				addrSelect <= eAddrSP;	
				dataSelect <= eDataPCHi;
				we <= 1'b0;
				spFunc <= fDecrement;
				nextState <= sCall1;
			end
			sCall1: begin
				// addrsel=sp, dbus <= pclo, sp <= sp-1
				eaFunc <= 2'bxx;
				addrSelect <= eAddrSP;	
				dataSelect <= eDataPCLo;
				we <= 1'b0;
				spFunc <= fDecrement;
				nextState <= sJump0;		
			end
			sReturn0: begin
				// addrsel=sp, ealo <= dbus, sp <= sp+1
				addrSelect <= eAddrSP;	
				eaFunc <= fLoadLo;
				spFunc <= fIncrement;
				pcFunc <= 2'bxx; // PC is about to be overwritten, so don't care
				nextState <= sReturn1;
			end
			sReturn1: begin
				// addrsel=sp, pc <= {dbus,ealo}
				eaFunc <= 2'bxx;
				addrSelect <= eAddrSP;
				pcFunc <= fLoad;
				nextState <= sSkipByte0;
			end
		endcase
	end	
	 
endmodule